# Writing charon

First: what is Charon?
Charon is a simple anti DDoS reverse proxy written in Rust, and it's being used to protect this website _right now_.

A DDoS protection program should embody three fundamental aspects:
- DDoS mitigation
- Low overhead
- End-user convenience

To achieve these core principles, I need to prioritise my program's performance, get it out of the way of well-intentioned users and simultaneously mitigate the server's susceptibility to a DDoS attack.

## But first, what is a DDoS attack?

A DDoS attack is a coordinated, distributed overloading of a server's systems. An army of weaponized computers all concurrently request data from a server, undermining its ability to respond to real clients. Eventually, the computational stress from this perpetual backlog of requests will exhaust the server's resources, causing it to crash.

There are a plethora of ways a server could mitigate the risk of DDoS attacks, but a few intuitive methods stand out, one of which is load balancing.

Load balancing involves the distribution of incoming requests from clients to an array of servers. Effective load balancers can monitor and anticipate the current computational load of partipating servers and divert requests accordingly. The problem with load balancing is that it requires multiple servers, and you must scale the number of your servers as attacks increase in volume and frequency.

Our next option doesn't require extra computing power, and oftentimes is free. Cloudflare is an American tech giant that protects its customers from DDoS attacks by lending them its insurmountable computing/network resources. Basically, Cloudflare acts as a bouncer in between a company's server, and a client. Using many arguably invasive tactics, it asseses the client's trustworthiness and subsequently either discards the client's request, allows it to access the server, or, in the case that the bouncer's unsure, makes the client solve a puzzle to verify its trustworthiness. The main concern with Cloudflare lies in the fact that it intercepts client requests. Like a bouncer, it could, in theory, keep a lengthy record of every website and its visitors. While that data is already known to your ISP, Cloudflare could also potentially view a visitor's activity on a website. According to vice.com, one of Cloudflare's co-founders, Mr Prince, asserts that the value of Cloudflare isn't in their code, but more in the data it collects. As of 2024, Cloudflare intercepts the traffic of half of the top 10,000 most popular websites [[reference]](https://backlinko.com/cloudflare-users).

From an internal perspective, Cloudflare's value lies in its data collection, but from an enterprise perspective, its utility starts and ends with its DDoS protection. My program, Charon has a few distinctions from Cloudflare's DDoS solution:
- DDoS protection method

    In contrast to how Cloudflare blocks any suspicious traffic, Charon takes a transactional approach to DDoS protection. When a new client tries to access your website, Charon presents the client with a PoW (Proof-of-Work) challenge. Solving this challenge takes an amount of computing power non-trivial to regular clients, and takes around 5 seconds to solve depending on the configuration. But if a malicious actor were to try and send many concurrent requests to the website, the cost of computing power would grow exponentially, rendering the attack financially unviable. This means that while Cloudflare automatically blocks TOR exit nodes, Charon can allow traffic from any source, regardless of their reputation.
- Self-hosted

    In terms of trustworthiness, Charon beats Cloudflare because of the fact that it is self hosted. You are intercepting your own traffic, which means the user activity on your site lies only on your and your clients' computers.

With that out of the way, I'll get into how I wrote Charon.

## Building a stack

> (In programming a "stack" is a set of softwares that work in tandem to provide an end product.)

Web-facing programs have two primary components - a frontend, and a backend. A frontend is the side of the service shown to the end-users, and oftentimes displays dynamic data. The backend handles the constuction and retrieval of said dynamic data, and also handles most of the logic of the program. To maximise the performance of my program, I'll need to optimise both the frontend and the backend.

As with any programming project, I'll need to select a language. For this project, to emphasize performance and developer experience, I've selected Rust. Rust is a highly performant language, rivalling the speed of the perennial C. It also has a massive ecosystem, with libraries that frictionlessly streamline programming.

### Templating libraries

A templating library is a way for the backend of a program to relay data to the end user. Say I had a function, `visitor_count` that counted the total number of visitors to my website. Using a templating library, I can easily interpolate that value into the HTML that is sent to the client:
```html
<h1> Lorem Ipsum </h1>
<p>Hello, world! You are the {visitor_count()}th viewer of this page!</p>
```

A templating library must 'lex' HTML, which is to say it must be able to form and manipulate an organised model of all the elements of an HTML page. Using that data, it must know exactly where and how to interpolate different types of data from the backend and render the resultant HTML to the end user.

It's imperative to the performance of a program to use an efficient templating engine, especially because it's in charge of so many moving parts.


```
cmp_compile/handlebars  time:   [47.043 µs 47.163 µs 47.295 µs]
cmp_compile/liquid      time:   [28.669 µs 28.776 µs 28.919 µs]
cmp_compile/minijinja   time:   [4.6921 µs 4.7003 µs 4.7092 µs]
cmp_compile/tera        time:   [35.145 µs 35.227 µs 35.315 µs]

cmp_render/askama       time:   [1.5744 µs 1.5793 µs 1.5846 µs]
cmp_render/handlebars   time:   [6.3056 µs 6.3221 µs 6.3399 µs]
cmp_render/liquid       time:   [11.364 µs 11.394 µs 11.426 µs]
cmp_render/minijinja    time:   [5.0475 µs 5.0600 µs 5.0739 µs]
cmp_render/tera         time:   [7.1101 µs 7.1285 µs 7.1482 µs]
```

Benchmarks recorded by templating library project _[minijinja](https://github.com/mitsuhiko/minijinja/tree/main/benchmarks)_ show that Askama is the fastest templating library to render. Since I'm prioritising user-experience, compile time isn't a concern in my project.

### HTTP libraries

An HTTP library facilitates the communication of a program and a client over the Internet. In the Rust ecosystem, there is a wide range of HTTP libraries of varying abstraction and use-case. To prioritise performance, I'd like my program to support the fewest features possible, while still fulfilling its purpose, which is why I chose [hyper](https://hyper.rs). Hyper is a relatively unabstracted HTTP library that allows me to handle exactly how my server responds to incoming requests.

And with that, I've created a stack for my program to use: Hyper and Askama.

## Proof of Work

As stated previously, the core of charon's anti-DDoS function is its Proof of Work system - a way for the client to prove to the server that its completed a computationally expensive task. Let me introduce the concept of an intractible function:

An intractible function is a function that is easy to do one way, but really hard to invert. Imagine a room with a pool player and a pool table with a single cue ball in the middle. Suppose you left this room, came back in 2 hours and discovered the cue ball was in a completely different place. Is it possible to find the exact number of times the pool player struck the cue ball? In this example, the intractible function is the act of repeatedly hitting the cue ball.

A hash is a type of intractible function used to verify the integrity of data. When given a certain input, a hash function will always generate the same output. But, making even a small change to the input will make the resulting output function completely unrecogniseable. In its essence, a hash algorithm takes the data you give it and precisely digests it into a completely different form using a pipeline of intractible functions.

Hashing the word 'hello' gives us the digest `2cf24dba5fb0a30e26e83b2ac5b9e29e1b161e5c1fa7425e73043362938b9824`. A key fundamental of hashes is the idea that it's extremely hard to find the original data from a hash. So, finding the word 'hello' from `2cf24dba5fb0a30e26e83b2ac5b9e29e1b161e5c1fa7425e73043362938b9824` is near impossible.

Using this idea, we can create a Proof of Work function that requires the client to find the original form of a hash. "But you just said finding the original form of a hash is near-impossible!" You're right, so we'll have to impose a few restrictions on the original data.

If I told you to find the original form of the hash: `4e07408562bedb8b60ce05c1decfe3ad16b72230967de01f640b7e4729b49fce`, you'd give up immediately because brute-forcing it could take thousands of years. But, if I supplied you with the information that the original data is a number between 1 and 10, you could find the original data simply by hashing all the possible inputs.

1 -> `6b86b273ff34fce19d6b804eff5a3f5747ada4eaa22f1d49c01e52ddb7875b4b`
2 -> `d4735e3a265e16eee03f59718b9b5d03019c07d8b6c51f90da3a666eec13ab35`
3 -> `4e07408562bedb8b60ce05c1decfe3ad16b72230967de01f640b7e4729b49fce` Congrats! You found it!

Using this logic, charon hashes a number from 1-200,000 and gives your computer that number's hash and the information that the original input is somewhere between 1 and 200,000. It then instructs your computer to find the number through brute force. The average household computer can compute 20,000 hashes per second, meaning, at worst, you'd have to wait about 10 seconds to enter the site. And while incredibly ingenious, this approach has its own vulnerabilities.

If you'd known beforehand that we'd play the 1-10 guessing game and you wanted to crack as fast as possible, you'd probably be driven to use a table. What I mean by this, is that you'd keep a record of the hashes from 1-10, so you don't have to compute them by demand. Similarly, a bad actor could precompute the hashes of the numbers 1-200,000 and instantly get through our proof of work system.

To extinguish this issue, I've implemented something called a 'salt'. Instead of computing the hash of a number by itself, charon prepends a randomly generated string to the number, like this: `fNCLXAwC3pCRIo3U` `28347`. Since the salt is different every time, your computer needs to manually compute the hashes from `fNCLXAwC3pCRIo3U` `0` to `fNCLXAwC3pCRIo3U` `200000`.

## Implementation

On startup, Charon starts two servers. A gateway server, and a challenge server:
```
 INFO  charon::gateway > Listening on http://127.0.0.1:8080
 INFO  charon::gateway > Proxying http://127.0.0.1:8081
 INFO  charon::challenge > Challenge server running at http://127.0.0.1:27764
```
The gateway server listens for incoming requests at port 8080, and is responsible for diverting authenticated traffic to the actual server at port 8081, and unauthenticated traffic to the challenge server at a random port, in this case, 27764.

When a client accesses the website for the first time, they're met with this:


![Charon please wait webpage](/assets/screenie.png "Screenshot of Charon")

Let's look at what happened under the hood. First, charon's gateway server received an unauthenticated request and forwarded it to the challenge server. The challenge server received the request and generated the proof of work problem to relay back to the client. Upon solving it, the client sets a cookie `__ch_auth` to be equal to its solution and retries the request. The gateway server receives an authenticated request and checks if the value of `__ch_auth` is valid. If it is, it forwards the request to the actual server, giving us this:

![Homepage](/assets/home_screenie.png "Screenshot of home page")

Thanks for reading this blog post. Charon's usage and source code can be found [here](http://git.fwfwqtpi2ofmehzdxe3e2htqfmhwfciwivpnsztv7dvpuamhr72ktlqd.onion/fancy/charon).