use askama::Template;
use log::debug;
use serde::{Deserialize, Serialize};
use std::{
    ffi::OsStr,
    fs::{self, File},
    io::Write,
    path::PathBuf,
    sync::{atomic::AtomicUsize, Arc},
};
use templates::BlogPage;
use tide::{http::mime, Request};
use tide_serve_dir_macro::auto_serve_dir;

#[derive(Debug, Serialize, Deserialize)]
struct BlogPost {
    pub title: String,
    pub description: String,
    pub id: String,
}

mod middlewarez;
mod templates;

#[derive(Debug, Clone)]
struct State {
    pub visitors: Arc<AtomicUsize>,
}

#[async_std::main]
async fn main() -> Result<(), Box<dyn std::error::Error>> {
    {
        fern::Dispatch::new()
            // Perform allocation-free log formatting
            .format(move |out, message, record| {
                out.finish(format_args!(
                    "[{} {} {}] {}",
                    humantime::format_rfc3339_millis(std::time::SystemTime::now()),
                    record.level(),
                    record.target(),
                    message
                ))
            })
            .level(log::LevelFilter::Debug)
            .chain(std::io::stderr())
            .chain(fern::log_file(".lethe.log")?)
            .apply()?;
    }
    let mut app = tide::with_state(State {
        visitors: Arc::new(AtomicUsize::new(
            fs::read_to_string("visitors.count")
                .unwrap_or_else(|_| "0".to_string())
                .parse()?,
        )),
    });
    app.with(tide_compress::CompressMiddleware::new());
    app.with(middlewarez::LoggerMiddleware::new());
    app.with(middlewarez::MinifyMiddleware::new());

    app.at("/").get(index);
    app.at("/blog").get(blog);
    app.at("/blog/:id").get(blog_post);
    auto_serve_dir!(app, "/assets", "assets");
    app.listen("127.0.0.1:8081").await?;
    Ok(())
}

async fn index(req: Request<State>) -> tide::Result {
    let state = req.state().clone();
    let visitors = state
        .visitors
        .fetch_add(1, std::sync::atomic::Ordering::Relaxed);
    if visitors % 10 == 0 {
        {
            use std::fs::OpenOptions;
            let mut file = OpenOptions::new()
                .write(true)
                .truncate(true)
                .create(true)
                .open("visitors.count")?;
            file.write_all(&visitors.to_string().as_bytes())?;
        }
    }
    let res = tide::Response::builder(200)
        .body(templates::HomePage { visitors }.render().unwrap())
        .content_type(mime::HTML)
        .build();
    return Ok(res);
}

async fn blog(_: Request<State>) -> tide::Result {
    let mut posts = Vec::new();
    for post in fs::read_dir("blog")? {
        let post = post?;
        if post.path().extension().unwrap() != OsStr::new("json") {
            continue;
        }
        let post: BlogPost = serde_json::from_reader(File::open(post.path()).unwrap())?;
        posts.push(post);
    }
    posts.reverse();
    let res = tide::Response::builder(200)
        .body(BlogPage { posts }.render().unwrap())
        .content_type(mime::HTML)
        .build();
    return Ok(res);
}

async fn blog_post(req: Request<State>) -> tide::Result {
    let post_path = req.param("id")?;
    let md_path = format!("blog/{post_path}.md");
    debug!("rendering {}..", md_path);
    let content = markdown::file_to_html(&PathBuf::from(md_path))?;
    let res = tide::Response::builder(200)
        .body(content)
        .content_type(mime::HTML)
        .build();
    return Ok(res);
}
